#!/usr/bin/python

# This is a simple non-RPC test case which runs a loop
# in a background thread and then inserts a debug hook

import threading
import time
from rpcpdb import terminal, UPdb_mixin


class PrimeServer(UPdb_mixin):
    def next_prime(self, p):
        """
        Naive, flaky and slow way to get the next
        prime number after that given.
        """
        if p < 2:
            # return first prime: 2
            return 2
        i = p  # start where we left off
        while True:
            i += 1
            probe = 2
            while True:
                if i % probe == 0:
                    # i is composite
                    break
                if probe >= p:
                    # i is co-prime to integers up to p.
                    return i
                probe += 1

    def test_func(self, **k):
        if k.pop('thing', None) == 'hello':
            print("Hiya")
        else:
            print("OK then.")

    def mainloop(self):
        p = 0
        while p < 150:
            p = self.next_prime(p)
            print(p)
            time.sleep(0.1)


def main():
    ps = PrimeServer()
    # test basic parameter criteria match
    t = threading.Thread(target=ps.mainloop)
    t.daemon = True
    t.start()
    time.sleep(2)
    terminal(ps.debug_func('next_prime',
                           match_criteria={'p': 97}))
    terminal(ps.debug_func('next_prime',
                           ignore_count=5))
    t.join()

    # test keyword criteria match
    def _():
        time.sleep(2)
        ps.test_func()
        time.sleep(2)
        ps.test_func(thing='hello')
    t = threading.Thread(target=_)
    t.daemon = True
    t.start()
    terminal(ps.debug_func('test_func',
                           match_criteria={'thing': 'hello'}))
    terminal(ps.debug_func('test_func',
                           match_criteria={'thing': 'hello'}))
    time.sleep(1)
    t.join()

if __name__ == '__main__':
    main()



#!/usr/bin/env python -u

from SimpleXMLRPCServer import SimpleXMLRPCServer
from SimpleXMLRPCServer import SimpleXMLRPCRequestHandler
from SocketServer import ThreadingMixIn


# Restrict to a particular path.
class RequestHandler(SimpleXMLRPCRequestHandler):
    rpc_paths = ('/RPC2',)


# Threaded mix-in
class ThreadedXMLRPCServer(ThreadingMixIn,
                           SimpleXMLRPCServer):
    pass

# Create server
server = ThreadedXMLRPCServer(("localhost", 8000),
                              requestHandler=RequestHandler,
                              allow_none=True)
server.register_introspection_functions()


class RpcServer(object):
    def next_prime(self, p):
        """
        Naive, flaky and slow way to get the next
        prime number after that given.
        """
        if p < 2:
            # return first prime: 2
            return 2
        i = p  # start where we left off
        while True:
            i += 1
            probe = 2
            while True:
                if i % probe == 0:
                    # i is composite
                    break
                if probe >= p:
                    # i is co-prime to integers up to p.
                    return i
                probe += 1

if __debug__:
    import updb

    class DebuggableRpcServer(RpcServer, updb.UPdb_mixin):
        pass

    server_instance = DebuggableRpcServer()
else:
    server_instance = RpcServer()

server.register_instance(server_instance)
# Run the server's main loop
server.serve_forever()



#!/usr/bin/env python -u

import Pyro.core
from rpcpdb import cli


def get_api_connection():
    return Pyro.core.getProxyForURI("PYROLOC://localhost:7766/rpc")
cli.get_api_connection = get_api_connection


if __name__ == '__main__':
    cli.main()



#!/usr/bin/python -u
"""
updb.py - the core of rpcpdb.

Contains mixin methods for RPC servers to support client-initiated
function debugging & tracing

Copyright 2010-2013 Ben Bass <benbass@codedstructure.net>
Licence: MIT - http://www.opensource.org/licenses/mit-license.php
"""

try:
    from inspect import getcallargs, getargspec
except ImportError:
    # more support for older (2.x) Python versions...
    from rpcpdb.inspect_helper import getcallargs, getargspec  # NOQA

import pdb
import socket
import sys
import os
import tempfile
import threading
import code

from rpcpdb.func_trace import Tracer


class UPdb(pdb.Pdb):
    def __init__(self, sock_path=None, level=0, force=False):
        if sock_path is None:
            # create a random socket - this can be
            # retrieved by self.debug_socket prior to
            # entering a Pdb session
            self._sock_path = os.path.join(tempfile.mkdtemp(),
                                           str(os.getpid()))
        elif force:
            try:
                os.remove(sock_path)
            except OSError:
                pass
        self._sock_path = sock_path
        self._level = level
        self._sock = socket.socket(socket.AF_UNIX,
                                   socket.SOCK_STREAM)
        self._sock.bind(self._sock_path)
        self._sock.listen(1)
        self._conn = self._sock.accept()[0]
        self._handle = self._conn.makefile('rw')
        self._old_stdin = sys.stdin
        self._old_stdout = sys.stdout
        extra = {}
        if 'nosigint' in getargspec(pdb.Pdb.__init__)[0]:
            # Python3.2 for some reason sets a SIGINT handler
            # when '(c)ontinue' is run (to allow resuming debugger
            # with Ctrl-C), but that fails when run in non-main thread,
            # so turn it off.
            extra = {'nosigint': True}
        pdb.Pdb.__init__(self,
                         stdin=self._handle,
                         stdout=self._handle,
                         **extra)
        sys.stdout = sys.stdin = self._handle

    @property
    def debug_socket(self):
        """read-only access to the debug socket path"""
        return self._sock_path

    def __enter__(self):
        target_frame = sys._getframe().f_back
        for count in range(self._level):
            try:
                target_frame = target_frame.f_back
            except AttributeError:
                break
        self.set_trace(target_frame)

    def __exit__(self, exc_type, exc_value, tb):
        sys.stdin = self._old_stdin
        sys.stdout = self._old_stdout

        self._conn.shutdown(socket.SHUT_RDWR)
        os.remove(self._sock_path)
        os.rmdir(os.path.dirname(self._sock_path))


class DebugConsole(code.InteractiveConsole):

    def __init__(self, socket_addr, *o, **k):
        code.InteractiveConsole.__init__(self, *o, **k)
        self.__sock_path = socket_addr
        self.__sock = socket.socket(socket.AF_UNIX,
                                    socket.SOCK_STREAM)
        self.__sock.bind(self.__sock_path)
        self.__sock.listen(1)
        self.__conn = self.__sock.accept()[0]
        self._debug_handle = self.__conn.makefile('rw')
        #sys.stdout = self._debug_handle
        #sys.stderr = self._debug_handle
        #sys.stdin = self._debug_handle
        sys.displayhook = self._displayhook

    def _displayhook(self, obj):
        if obj is None:
            return
        #__builtins__._ = None
        self.write(repr(obj) + '\n')
        #__builtins__._ = obj

    def write(self, data):
        if data:
            self._debug_handle.write(data)
            self._debug_handle.flush()

    def raw_input(self, prompt=''):
        self.write(prompt)
        return self._debug_handle.readline()


class UPdb_mixin(object):
    """
    This is designed to be mixed-in to an RPC server
    object. It allows an RPC client to register a
    RPC method as being debuggable.

    On triggering, subsequent calls to that method
    will trigger a pdb breakpoint (currently with
    interaction over a unix socket)
    """
    # these are maps from function name to
    # (function, unix_socket_path/trace_path) pairs.
    _updb_debug_func = {}
    _updb_trace_func = {}

    def debug_func(self, f, once=True, force=True,
                   match_criteria=None, ignore_count=0):
        """
        @param f: function name
        @param once: one-shot debug - undebug after first hit
        @param force: override any existing unix socket
        @param match_criteria: dictionary of param name/value matches
        @param ignore_count: don't trigger for ignore_count matches
        @return: path to debug socket
        """
        if f in self._updb_debug_func:
            return self._updb_debug_func[f][1]
        else:
            func = getattr(self, f)
            pdb_sock_path = "%s/pdb_sock" % tempfile.mkdtemp(prefix='updb_')
            self._updb_debug_func[f] = (func, pdb_sock_path)

            def arg_match(o, k):
                # check whether function f is called with
                # parameters matching the given match_criteria
                if not match_criteria:
                    # we always match if we have no match criteria
                    return True
                called_with = getcallargs(func, *o, **k)
                # insert anything passed using **kwargs etc as if it were
                # a standard named parameter (for the purposes of comparison)
                fn_keyword_arg_name = getargspec(func)[2]
                if fn_keyword_arg_name is not None:
                    for kw_k, kw_v in called_with.get(fn_keyword_arg_name,
                                                      {}).items():
                        called_with[kw_k] = kw_v

                match = True
                # All things in match_criteria must have exact matches
                # in called_with for us to match things.
                for key, val in match_criteria.items():
                    # note short-circuit logic here
                    if key not in called_with or called_with[key] != val:
                        match = False

                return match

            def debug_check(*o, **k):
                if arg_match(o, k):
                    if debug_check._ignore_count <= 0:
                        if once:
                            self.undebug_func(f)
                        with UPdb(pdb_sock_path, level=1, force=force):
                            return func(*o, **k)
                    debug_check._ignore_count -= 1
                # we're not debugging you, this time...
                return func(*o, **k)
            # use a function attribute to store the skip count;
            # decrement this on each matched call. Trigger at <= 0
            debug_check._ignore_count = ignore_count

            setattr(self, f, debug_check)
            return pdb_sock_path

    def undebug_func(self, f):
        """
        don't debug this function anymore.
        """
        if f in self._updb_debug_func:
            setattr(self, f, self._updb_debug_func.pop(f)[0])

    def trace_func(self, f, tracepath, once=True,
                   match_criteria=None, ignore_count=0):
        """
        @param f: function name
        @param tracepath: pathname to write trace to
        @param once: one-shot debug - untrace after first hit
        @param match_criteria: dictionary of param name/value matches
        @param ignore_count: don't trigger for ignore_count matches
        @return: path to trace file
        """
        if f in self._updb_trace_func:
            return self._updb_trace_func[f][1]
        else:
            func = getattr(self, f)
            self._updb_trace_func[f] = (func, tracepath)

            def arg_match(o, k):
                # check whether function f is called with
                # parameters matching the given match_criteria
                if not match_criteria:
                    # we always match if we have no match criteria
                    return True
                called_with = getcallargs(func, *o, **k)
                # insert anything passed using **kwargs etc as if it were
                # a standard named parameter (for the purposes of comparison)
                fn_keyword_arg_name = getargspec(func)[2]
                if fn_keyword_arg_name is not None:
                    for kw_k, kw_v in called_with.get(fn_keyword_arg_name,
                                                      {}).items():
                        called_with[kw_k] = kw_v

                match = True
                # All things in match_criteria must have exact matches
                # in called_with for us to match things.
                for key, val in match_criteria.items():
                    # note short-circuit logic here
                    if key not in called_with or called_with[key] != val:
                        match = False

                return match

            def debug_check(*o, **k):
                if arg_match(o, k):
                    if debug_check._ignore_count <= 0:
                        if once:
                            self.untrace_func(f)
                        old_tracer = sys.gettrace()
                        sys.settrace(Tracer(tracepath))
                        try:
                            return func(*o, **k)
                        finally:
                            sys.settrace(old_tracer)
                    debug_check._ignore_count -= 1
                # we're not tracing you, this time...
                return func(*o, **k)
            # use a function attribute to store the skip count;
            # decrement this on each matched call. Trigger at <= 0
            debug_check._ignore_count = ignore_count

            setattr(self, f, debug_check)

    def untrace_func(self, f):
        """
        don't trace this function anymore.
        """
        if f in self._updb_trace_func:
            setattr(self, f, self._updb_trace_func.pop(f)[0])

    def debug_interactive_console(self):
        """
        start an interactive console in a new thread
        return socket address to connect to it.
        """
        con_sock_path = "%s/con_sock" % tempfile.mkdtemp(prefix='updb_')
        context = globals()
        context['_debug_api'] = self

        def run_console():
            context['_debug_console'] = DebugConsole(con_sock_path, context)
            try:
                context['_debug_console'].interact()
            finally:
                os.remove(con_sock_path)
                os.rmdir(os.path.dirname(con_sock_path))

        t = threading.Thread(target=run_console)
        t.daemon = True
        t.start()
        return con_sock_path



#!/usr/bin/env python -u

import xmlrpclib
from rpcpdb import cli


def get_api_connection():
    return xmlrpclib.ServerProxy('http://localhost:8000')
cli.get_api_connection = get_api_connection


if __name__ == '__main__':
    cli.main()



#!/usr/bin/python -u

"""
Command line interface to rpcpdb. This becomes an RPC client allowing
RPC methods to be debugged & traced.

Copyright 2010-2013 Ben Bass <benbass@codedstructure.net>
Licence: MIT - http://www.opensource.org/licenses/mit-license.php
"""

import os
import ast
import signal

from rpcpdb import termsock


def get_api_connection():
    """
    replace this with something to return a proxy to your API
    """
    raise NotImplementedError('this function should be replaced')


def debug(options):
    if options.funcname:
        # get a client connection
        c = get_api_connection()
        match_criteria = {}
        for match in (options.match or []):
            param, value = [x.strip() for x in match.split(':', 1)]
            if not param and not value:
                raise ValueError('invalid match definition "%s"' % match)
            try:
                real_value = ast.literal_eval(value)
            except SyntaxError:
                raise ValueError('invalid match definition "%s"' % match)
            except ValueError:
                # literal_eval gives a ValueError where I would expect
                # a NameError...
                real_value = value
            match_criteria[param] = real_value
        if options.trace:
            c.trace_func(options.funcname,
                         options.trace_logfile,
                         once=options.once,
                         match_criteria=match_criteria,
                         ignore_count=options.ignore_count)
            sock_name = None
        else:
            sock_name = c.debug_func(options.funcname,
                                     once=options.once,
                                     force=options.force,
                                     match_criteria=match_criteria,
                                     ignore_count=options.ignore_count)

        def _cleanup(*args):
            # This undoes the breakpoint in the event we are
            # rudely interrupted.
            c.undebug_func(options.funcname)
            raise SystemExit('Aborted.')

        signal.signal(signal.SIGINT, _cleanup)
        signal.signal(signal.SIGTERM, _cleanup)

    elif options.console:
        c = get_api_connection()
        sock_name = c.debug_interactive_console()
    else:
        sock_name = options.sockname

    if sock_name is not None:
        dbg = termsock.TermSock(sock_name)
        dbg.mainloop()


def main():
    import optparse
    usage = "%prog: connect to a UPdb debugger"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("-s", "--sockname",
                      dest="sockname",
                      help="UNIX-domain socket path")
    parser.add_option("-a", "--funcname",
                      dest="funcname",
                      help="API function to debug")
    parser.add_option("-c", "--console",
                      dest="console",
                      action="store_true",
                      help="Interactive console")
    parser.add_option("-m", "--match",
                      dest="match",
                      action="append",
                      help="Add a match criteria in form 'param:value'")
    parser.add_option("-o", "--once",
                      dest="once",
                      action="store_false",
                      help="Set breakpoint as once-only (default)",
                      default=True)
    parser.add_option("-i", "--ignore_count",
                      dest="ignore_count",
                      default=0,
                      type=int,
                      help="Ignore count (default 0)")
    parser.add_option("-f", "--force",
                      dest="force",
                      action="store_true",
                      help="Force (overwrite any previous socket)")
    parser.add_option("-t", "--trace",
                      dest="trace",
                      action="store_true",
                      help="Trace (rather than debug) function call")
    parser.add_option("-l", "--trace-logfile",
                      dest="trace_logfile",
                      default=os.path.abspath("trace.out"),
                      help="Path to trace file to store (default 'trace.out')")
    options, args = parser.parse_args()
    if len(filter(None, (options.sockname,
                         options.funcname,
                         options.console))) != 1:
        parser.error("Specify exactly one of --funcname, --sockname or --console")
    if (any((options.match, options.ignore_count, options.trace)) and
        not options.funcname):
        parser.error("options requiring --funcname given")
    if options.force and options.trace:
        parser.error("--force can not be specified with --trace")

    debug(options)


if __name__ == '__main__':
    main()



#!/usr/bin/python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name="rpcpdb",
    version="0.4.2",
    description="Debug support for RPC servers",
    long_description=open('README.rst').read(),
    author="Ben Bass",
    author_email="benbass@codedstructure.net",
    url="http://bitbucket.org/codedstructure/rpcpdb",
    packages=["rpcpdb"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Topic :: Software Development :: Debuggers",
    ]
)



"""
Tracing of functions, writing trace files.

Copyright 2013 Ben Bass <benbass@codedstructure.net>
Licence: MIT - http://www.opensource.org/licenses/mit-license.php
"""


import os
import sys
import inspect


class Tracer(object):
    """
    set this with sys.settrace() to perform tracing
    """
    def __init__(self, tracepath=None):
        """
        @param tracepath: file object to write trace log to
        """
        if tracepath is None:
            self.tracefile = sys.stdout
        else:
            self.tracefile = open(tracepath, 'w')

    def __call__(self, frame, event, arg):
        """
        top-level trace function

        see sys.settrace() documentation
        """
        if event in ('call',):
            return LocalTracer(frame, event, tracefile=self.tracefile)
        else:
            self.tracefile.write('>> %s ' % event)


class LocalTracer(object):
    """
    local trace class - this is instantiated by the global Tracer() class
    """

    def __init__(self, frame, event, tracefile=None):
        """
        @param frame, event - passed in from top-level tracer
        @param tracefile: file object to write trace log to
        """
        if tracefile is None:
            tracefile = sys.stdout
        self.tracefile = tracefile
        self.fname = inspect.getframeinfo(frame).function

        self.obj_source = None
        self.line_offset = None
        self._skipping = False
        try:
            self.obj_source, self.line_offset = inspect.getsourcelines(frame.f_code)
        except IOError:
            # source code unavailable
            pass
        else:
            self.max_line_len = min(80, len(max(self.obj_source, key=len)))
            self.fmt_str = "{}:{:<4} ||| {:%d} ||| {{{}}}\n" % self.max_line_len
            filename = frame.f_code.co_filename
            # only keep last two elements of filename path
            self.srcfile = os.path.sep.join(filename.split(os.path.sep)[-2:])

        if event == 'call':
            self.tracefile.write("=> {}\n".format(self.fname))

    def _shortstr(self, s):
        if len(repr(s)) > 80:
            return repr(s)[:80] + '...'
        else:
            return repr(s)

    def __call__(self, frame, event, arg):
        """
        local trace function, called on each line/call/return/exception etc
        within a local scope

        See sys.settrace() documentation
        """
        if event == 'line':
            if self.obj_source is None:
                if not self._skipping:
                    self.tracefile.write("--\n")
                    self._skipping = True
                return None

            lineno = frame.f_lineno
            try:
                sourceline = self.obj_source[lineno - self.line_offset].rstrip()
            except Exception:
                # perhaps something is out of sync or source doesn't exist
                if not self._skipping:
                    self.tracefile.write("---\n")
                    self._skipping = True
                return None
            else:
                self._skipping = False
                localvars = frame.f_locals.copy()
                # design decision - omit certain things here. There are
                # occasions when they will be useful & more cases where other
                # things are just clutter, but...
                names_to_omit = ['self', 'builtins']
                names_to_omit.extend([key for key in localvars if key.startswith('__')])
                for key in names_to_omit:
                    localvars.pop(key, None)
                for key, value in list(localvars.items()):
                    localvars[key] = self._shortstr(value)
                context = ', '.join(('{}: {}'.format(k, v) for k, v in localvars.items()))

                self.tracefile.write(self.fmt_str.format(self.srcfile, lineno,
                                                         sourceline, context))
            return self

        elif event == 'return':
            self.tracefile.write("<= ({})  # from: {}\n".format(self._shortstr(arg),
                                                                self.fname))

        elif event == 'exception':
            # support for exceptions
            self.tracefile.write('%s(%r)\n' % (arg[0].__name__, arg[1]))

        else:
            # This "shouldn't" happen...
            self.tracefile.write('>>> %s, %s\n' % (event, arg))

#
# The following may be helpful if using this module standalone as an example.
#
# Note this takes a long time on Python3, due to module initialisation in python
# Might want to turn off tracing of global-level module evaluation in future...
#


def testfunc():
    i = 0
    while i < 10:
        i = i * 2 + 1
        print (i)
    return i


def fact(i):
    if i == 0:
        return 1
    else:
        return i * fact(i - 1)


def reciprocal(i):
    return 1.0 / i


def test_exc():
    try:
        for i in range(3, -1, -1):
            z = reciprocal(i)
            print(z)
    except ZeroDivisionError:
        print("Done")


def test_c():
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('8.8.8.8', 53))
    print(s.getpeername())


def main():
    old_trace = sys.gettrace()
    sys.settrace(Tracer())
    try:
        fact(testfunc())
        test_exc()
        test_c()
    finally:
        sys.settrace(old_trace)


if __name__ == '__main__':
    main()



#!/usr/bin/python -u

import termsock
import Pyro.core


s = Pyro.core.getProxyForURI("PYROLOC://localhost:7766/rpc")
dbg_path = s.debug_func('next_prime', match_criteria={'p': 79})
dbg_path = s.debug_func('next_prime', ignore_count=2)

termsock.terminal(dbg_path)
